@@ -8,7 +8,7 @@
  * Darko Prenosil <Darko.Prenosil@finteh.hr>
  * Shridhar Daithankar <shridhar_daithankar@persistent.co.in>
  *
- * $PostgreSQL: pgsql/contrib/dblink/dblink.c,v 1.69.2.3 2010/02/03 23:01:34 joe Exp $
+ * $PostgreSQL: pgsql/contrib/dblink/dblink.c,v 1.69.2.8 2010/06/15 19:04:28 tgl Exp $
  * Copyright (c) 2001-2008, PostgreSQL Global Development Group
  * ALL RIGHTS RESERVED;
  *
@@ -52,6 +52,7 @@
 #include "nodes/nodes.h"
 #include "nodes/pg_list.h"
 #include "parser/parse_type.h"
+#include "parser/scansup.h"
 #include "tcop/tcopprot.h"
 #include "utils/acl.h"
 #include "utils/array.h"
@@ -82,19 +83,21 @@ static remoteConn *getConnectionByName(const char *name);
 static HTAB *createConnHash(void);
 static void createNewConnection(const char *name, remoteConn * rconn);
 static void deleteConnection(const char *name);
-static char **get_pkey_attnames(Oid relid, int16 *numatts);
+static char **get_pkey_attnames(Relation rel, int16 *numatts);
 static char **get_text_array_contents(ArrayType *array, int *numitems);
-static char *get_sql_insert(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pkattvals, char **tgt_pkattvals);
-static char *get_sql_delete(Oid relid, int2vector *pkattnums, int16 pknumatts, char **tgt_pkattvals);
-static char *get_sql_update(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pkattvals, char **tgt_pkattvals);
+static char *get_sql_insert(Relation rel, int *pkattnums, int pknumatts, char **src_pkattvals, char **tgt_pkattvals);
+static char *get_sql_delete(Relation rel, int *pkattnums, int pknumatts, char **tgt_pkattvals);
+static char *get_sql_update(Relation rel, int *pkattnums, int pknumatts, char **src_pkattvals, char **tgt_pkattvals);
 static char *quote_literal_cstr(char *rawstr);
 static char *quote_ident_cstr(char *rawstr);
-static int16 get_attnum_pk_pos(int2vector *pkattnums, int16 pknumatts, int16 key);
-static HeapTuple get_tuple_of_interest(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pkattvals);
-static Oid	get_relid_from_relname(text *relname_text);
-static char *generate_relation_name(Oid relid);
+static int	get_attnum_pk_pos(int *pkattnums, int pknumatts, int key);
+static HeapTuple get_tuple_of_interest(Relation rel, int *pkattnums, int pknumatts, char **src_pkattvals);
+static Relation get_rel_from_relname(text *relname_text, LOCKMODE lockmode, AclMode aclmode);
+static char *generate_relation_name(Relation rel);
 static void dblink_security_check(PGconn *conn, remoteConn *rconn);
-static int get_nondropped_natts(Oid relid);
+static void validate_pkattnums(Relation rel,
+				   int2vector *pkattnums_arg, int32 pknumatts_arg,
+				   int **pkattnums, int *pknumatts);
 
 /* Global */
 static remoteConn *pconn = NULL;
@@ -1139,20 +1142,23 @@ PG_FUNCTION_INFO_V1(dblink_exec);
 Datum
 dblink_exec(PG_FUNCTION_ARGS)
 {
+	text	   *volatile sql_cmd_status = NULL;
+	PGconn	   *volatile conn = NULL;
+	volatile bool freeconn = false;
+
+	DBLINK_INIT;
+
+	PG_TRY();
+	{
 	char	   *msg;
 	PGresult   *res = NULL;
-	text	   *sql_cmd_status = NULL;
 	TupleDesc	tupdesc = NULL;
-	PGconn	   *conn = NULL;
 	char	   *connstr = NULL;
 	char	   *sql = NULL;
 	char	   *conname = NULL;
 	remoteConn *rconn = NULL;
-	bool		freeconn = false;
 	bool		fail = true;	/* default to backward compatible behavior */
 
-	DBLINK_INIT;
-
 	if (PG_NARGS() == 3)
 	{
 		/* must be text,text,bool */
@@ -1231,8 +1237,17 @@ dblink_exec(PG_FUNCTION_ARGS)
 				(errcode(ERRCODE_S_R_E_PROHIBITED_SQL_STATEMENT_ATTEMPTED),
 				 errmsg("statement returning results not allowed")));
 	}
+	}
+	PG_CATCH();
+	{
+		/* if needed, close the connection to the database */
+		if (freeconn)
+			PQfinish(conn);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 
-	/* if needed, close the connection to the database and cleanup */
+	/* if needed, close the connection to the database */
 	if (freeconn)
 		PQfinish(conn);
 
@@ -1251,7 +1266,6 @@ Datum
 dblink_get_pkey(PG_FUNCTION_ARGS)
 {
 	int16		numatts;
-	Oid			relid;
 	char	  **results;
 	FuncCallContext *funcctx;
 	int32		call_cntr;
@@ -1262,7 +1276,8 @@ dblink_get_pkey(PG_FUNCTION_ARGS)
 	/* stuff done only on the first call of the function */
 	if (SRF_IS_FIRSTCALL())
 	{
-		TupleDesc	tupdesc = NULL;
+		Relation	rel;
+		TupleDesc	tupdesc;
 
 		/* create a function context for cross-call persistence */
 		funcctx = SRF_FIRSTCALL_INIT();
@@ -1272,13 +1287,13 @@ dblink_get_pkey(PG_FUNCTION_ARGS)
 		 */
 		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
 
-		/* convert relname to rel Oid */
-		relid = get_relid_from_relname(PG_GETARG_TEXT_P(0));
-		if (!OidIsValid(relid))
-			ereport(ERROR,
-					(errcode(ERRCODE_UNDEFINED_TABLE),
-					 errmsg("relation \"%s\" does not exist",
-							GET_STR(PG_GETARG_TEXT_P(0)))));
+		/* open target relation */
+		rel = get_rel_from_relname(PG_GETARG_TEXT_P(0), AccessShareLock, ACL_SELECT);
+
+		/* get the array of attnums */
+		results = get_pkey_attnames(rel, &numatts);
+
+		relation_close(rel, AccessShareLock);
 
 		/*
 		 * need a tuple descriptor representing one INT and one TEXT column
@@ -1296,9 +1311,6 @@ dblink_get_pkey(PG_FUNCTION_ARGS)
 		attinmeta = TupleDescGetAttInMetadata(tupdesc);
 		funcctx->attinmeta = attinmeta;
 
-		/* get an array of attnums */
-		results = get_pkey_attnames(relid, &numatts);
-
 		if ((results != NULL) && (numatts > 0))
 		{
 			funcctx->max_calls = numatts;
@@ -1381,53 +1393,29 @@ Datum
 dblink_build_sql_insert(PG_FUNCTION_ARGS)
 {
 	text	   *relname_text = PG_GETARG_TEXT_P(0);
-	int2vector *pkattnums = (int2vector *) PG_GETARG_POINTER(1);
-	int32		pknumatts_tmp = PG_GETARG_INT32(2);
+	int2vector *pkattnums_arg = (int2vector *) PG_GETARG_POINTER(1);
+	int32		pknumatts_arg = PG_GETARG_INT32(2);
 	ArrayType  *src_pkattvals_arry = PG_GETARG_ARRAYTYPE_P(3);
 	ArrayType  *tgt_pkattvals_arry = PG_GETARG_ARRAYTYPE_P(4);
-	Oid			relid;
-	int16		pknumatts = 0;
+	Relation	rel;
+	int		   *pkattnums;
+	int			pknumatts;
 	char	  **src_pkattvals;
 	char	  **tgt_pkattvals;
 	int			src_nitems;
 	int			tgt_nitems;
 	char	   *sql;
-	int			nondropped_natts;
 
 	/*
-	 * Convert relname to rel OID.
+	 * Open target relation.
 	 */
-	relid = get_relid_from_relname(relname_text);
-	if (!OidIsValid(relid))
-		ereport(ERROR,
-				(errcode(ERRCODE_UNDEFINED_TABLE),
-				 errmsg("relation \"%s\" does not exist",
-						GET_STR(relname_text))));
+	rel = get_rel_from_relname(relname_text, AccessShareLock, ACL_SELECT);
 
 	/*
-	 * There should be at least one key attribute
+	 * Process pkattnums argument.
 	 */
-	if (pknumatts_tmp <= 0)
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("number of key attributes must be > 0")));
-
-	if (pknumatts_tmp <= SHRT_MAX)
-		pknumatts = pknumatts_tmp;
-	else
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("input for number of primary key " \
-						"attributes too large")));
-
-	/*
-	 * ensure we don't ask for more pk attributes than we have
-	 * non-dropped columns
-	 */
-	nondropped_natts = get_nondropped_natts(relid);
-	if (pknumatts > nondropped_natts)
-		ereport(ERROR, (errcode(ERRCODE_SYNTAX_ERROR),
-				errmsg("number of primary key fields exceeds number of specified relation attributes")));
+	validate_pkattnums(rel, pkattnums_arg, pknumatts_arg,
+					   &pkattnums, &pknumatts);
 
 	/*
 	 * Source array is made up of key values that will be used to locate the
@@ -1462,7 +1450,12 @@ dblink_build_sql_insert(PG_FUNCTION_ARGS)
 	/*
 	 * Prep work is finally done. Go get the SQL string.
 	 */
-	sql = get_sql_insert(relid, pkattnums, pknumatts, src_pkattvals, tgt_pkattvals);
+	sql = get_sql_insert(rel, pkattnums, pknumatts, src_pkattvals, tgt_pkattvals);
+
+	/*
+	 * Now we can close the relation.
+	 */
+	relation_close(rel, AccessShareLock);
 
 	/*
 	 * And send it
@@ -1491,50 +1484,26 @@ Datum
 dblink_build_sql_delete(PG_FUNCTION_ARGS)
 {
 	text	   *relname_text = PG_GETARG_TEXT_P(0);
-	int2vector *pkattnums = (int2vector *) PG_GETARG_POINTER(1);
-	int32		pknumatts_tmp = PG_GETARG_INT32(2);
+	int2vector *pkattnums_arg = (int2vector *) PG_GETARG_POINTER(1);
+	int32		pknumatts_arg = PG_GETARG_INT32(2);
 	ArrayType  *tgt_pkattvals_arry = PG_GETARG_ARRAYTYPE_P(3);
-	int			nondropped_natts;
-	Oid			relid;
-	int16		pknumatts = 0;
+	Relation	rel;
+	int		   *pkattnums;
+	int			pknumatts;
 	char	  **tgt_pkattvals;
 	int			tgt_nitems;
 	char	   *sql;
 
 	/*
-	 * Convert relname to rel OID.
+	 * Open target relation.
 	 */
-	relid = get_relid_from_relname(relname_text);
-	if (!OidIsValid(relid))
-		ereport(ERROR,
-				(errcode(ERRCODE_UNDEFINED_TABLE),
-				 errmsg("relation \"%s\" does not exist",
-						GET_STR(relname_text))));
+	rel = get_rel_from_relname(relname_text, AccessShareLock, ACL_SELECT);
 
 	/*
-	 * There should be at least one key attribute
+	 * Process pkattnums argument.
 	 */
-	if (pknumatts_tmp <= 0)
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("number of key attributes must be > 0")));
-
-	if (pknumatts_tmp <= SHRT_MAX)
-		pknumatts = pknumatts_tmp;
-	else
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("input for number of primary key " \
-						"attributes too large")));
-
-	/*
-	 * ensure we don't ask for more pk attributes than we have
-	 * non-dropped columns
-	 */
-	nondropped_natts = get_nondropped_natts(relid);
-	if (pknumatts > nondropped_natts)
-		ereport(ERROR, (errcode(ERRCODE_SYNTAX_ERROR),
-				errmsg("number of primary key fields exceeds number of specified relation attributes")));
+	validate_pkattnums(rel, pkattnums_arg, pknumatts_arg,
+					   &pkattnums, &pknumatts);
 
 	/*
 	 * Target array is made up of key values that will be used to build the
@@ -1554,7 +1523,12 @@ dblink_build_sql_delete(PG_FUNCTION_ARGS)
 	/*
 	 * Prep work is finally done. Go get the SQL string.
 	 */
-	sql = get_sql_delete(relid, pkattnums, pknumatts, tgt_pkattvals);
+	sql = get_sql_delete(rel, pkattnums, pknumatts, tgt_pkattvals);
+
+	/*
+	 * Now we can close the relation.
+	 */
+	relation_close(rel, AccessShareLock);
 
 	/*
 	 * And send it
@@ -1587,13 +1561,13 @@ Datum
 dblink_build_sql_update(PG_FUNCTION_ARGS)
 {
 	text	   *relname_text = PG_GETARG_TEXT_P(0);
-	int2vector *pkattnums = (int2vector *) PG_GETARG_POINTER(1);
-	int32		pknumatts_tmp = PG_GETARG_INT32(2);
+	int2vector *pkattnums_arg = (int2vector *) PG_GETARG_POINTER(1);
+	int32		pknumatts_arg = PG_GETARG_INT32(2);
 	ArrayType  *src_pkattvals_arry = PG_GETARG_ARRAYTYPE_P(3);
 	ArrayType  *tgt_pkattvals_arry = PG_GETARG_ARRAYTYPE_P(4);
-	int			nondropped_natts;
-	Oid			relid;
-	int16		pknumatts = 0;
+	Relation	rel;
+	int		   *pkattnums;
+	int			pknumatts;
 	char	  **src_pkattvals;
 	char	  **tgt_pkattvals;
 	int			src_nitems;
@@ -1601,39 +1575,15 @@ dblink_build_sql_update(PG_FUNCTION_ARGS)
 	char	   *sql;
 
 	/*
-	 * Convert relname to rel OID.
+	 * Open target relation.
 	 */
-	relid = get_relid_from_relname(relname_text);
-	if (!OidIsValid(relid))
-		ereport(ERROR,
-				(errcode(ERRCODE_UNDEFINED_TABLE),
-				 errmsg("relation \"%s\" does not exist",
-						GET_STR(relname_text))));
+	rel = get_rel_from_relname(relname_text, AccessShareLock, ACL_SELECT);
 
 	/*
-	 * There should be one source array key values for each key attnum
+	 * Process pkattnums argument.
 	 */
-	if (pknumatts_tmp <= 0)
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("number of key attributes must be > 0")));
-
-	if (pknumatts_tmp <= SHRT_MAX)
-		pknumatts = pknumatts_tmp;
-	else
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("input for number of primary key " \
-						"attributes too large")));
-
-	/*
-	 * ensure we don't ask for more pk attributes than we have
-	 * non-dropped columns
-	 */
-	nondropped_natts = get_nondropped_natts(relid);
-	if (pknumatts > nondropped_natts)
-		ereport(ERROR, (errcode(ERRCODE_SYNTAX_ERROR),
-				errmsg("number of primary key fields exceeds number of specified relation attributes")));
+	validate_pkattnums(rel, pkattnums_arg, pknumatts_arg,
+					   &pkattnums, &pknumatts);
 
 	/*
 	 * Source array is made up of key values that will be used to locate the
@@ -1668,7 +1618,12 @@ dblink_build_sql_update(PG_FUNCTION_ARGS)
 	/*
 	 * Prep work is finally done. Go get the SQL string.
 	 */
-	sql = get_sql_update(relid, pkattnums, pknumatts, src_pkattvals, tgt_pkattvals);
+	sql = get_sql_update(rel, pkattnums, pknumatts, src_pkattvals, tgt_pkattvals);
+
+	/*
+	 * Now we can close the relation.
+	 */
+	relation_close(rel, AccessShareLock);
 
 	/*
 	 * And send it
@@ -1705,7 +1660,7 @@ dblink_current_query(PG_FUNCTION_ARGS)
  * Return NULL, and set numatts = 0, if no primary key exists.
  */
 static char **
-get_pkey_attnames(Oid relid, int16 *numatts)
+get_pkey_attnames(Relation rel, int16 *numatts)
 {
 	Relation	indexRelation;
 	ScanKeyData skey;
@@ -1713,22 +1668,11 @@ get_pkey_attnames(Oid relid, int16 *numatts)
 	HeapTuple	indexTuple;
 	int			i;
 	char	  **result = NULL;
-	Relation	rel;
 	TupleDesc	tupdesc;
-	AclResult	aclresult;
 
 	/* initialize numatts to 0 in case no primary key exists */
 	*numatts = 0;
 
-	/* open relation using relid, check permissions, get tupdesc */
-	rel = relation_open(relid, AccessShareLock);
-
-	aclresult = pg_class_aclcheck(RelationGetRelid(rel), GetUserId(),
-								  ACL_SELECT);
-	if (aclresult != ACLCHECK_OK)
-		aclcheck_error(aclresult, ACL_KIND_CLASS,
-					   RelationGetRelationName(rel));
-
 	tupdesc = rel->rd_att;
 
 	/* Prepare to scan pg_index for entries having indrelid = this rel. */
@@ -1736,7 +1680,7 @@ get_pkey_attnames(Oid relid, int16 *numatts)
 	ScanKeyInit(&skey,
 				Anum_pg_index_indrelid,
 				BTEqualStrategyNumber, F_OIDEQ,
-				ObjectIdGetDatum(relid));
+				ObjectIdGetDatum(RelationGetRelid(rel)));
 
 	scan = systable_beginscan(indexRelation, IndexIndrelidIndexId, true,
 							  SnapshotNow, 1, &skey);
@@ -1762,7 +1706,6 @@ get_pkey_attnames(Oid relid, int16 *numatts)
 
 	systable_endscan(scan);
 	heap_close(indexRelation, AccessShareLock);
-	relation_close(rel, AccessShareLock);
 
 	return result;
 }
@@ -1829,32 +1772,27 @@ get_text_array_contents(ArrayType *array, int *numitems)
 }
 
 static char *
-get_sql_insert(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pkattvals, char **tgt_pkattvals)
+get_sql_insert(Relation rel, int *pkattnums, int pknumatts, char **src_pkattvals, char **tgt_pkattvals)
 {
-	Relation	rel;
 	char	   *relname;
 	HeapTuple	tuple;
 	TupleDesc	tupdesc;
 	int			natts;
 	StringInfoData buf;
 	char	   *val;
-	int16		key;
+	int			key;
 	int			i;
 	bool		needComma;
 
 	initStringInfo(&buf);
 
 	/* get relation name including any needed schema prefix and quoting */
-	relname = generate_relation_name(relid);
+	relname = generate_relation_name(rel);
 
-	/*
-	 * Open relation using relid
-	 */
-	rel = relation_open(relid, AccessShareLock);
 	tupdesc = rel->rd_att;
 	natts = tupdesc->natts;
 
-	tuple = get_tuple_of_interest(relid, pkattnums, pknumatts, src_pkattvals);
+	tuple = get_tuple_of_interest(rel, pkattnums, pknumatts, src_pkattvals);
 	if (!tuple)
 		ereport(ERROR,
 				(errcode(ERRCODE_CARDINALITY_VIOLATION),
@@ -1879,7 +1817,7 @@ get_sql_insert(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pka
 	appendStringInfo(&buf, ") VALUES(");
 
 	/*
-	 * remember attvals are 1 based
+	 * Note: i is physical column number (counting from 0).
 	 */
 	needComma = false;
 	for (i = 0; i < natts; i++)
@@ -1890,12 +1828,9 @@ get_sql_insert(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pka
 		if (needComma)
 			appendStringInfo(&buf, ",");
 
-		if (tgt_pkattvals != NULL)
-			key = get_attnum_pk_pos(pkattnums, pknumatts, i + 1);
-		else
-			key = -1;
+		key = get_attnum_pk_pos(pkattnums, pknumatts, i);
 
-		if (key > -1)
+		if (key >= 0)
 			val = tgt_pkattvals[key] ? pstrdup(tgt_pkattvals[key]) : NULL;
 		else
 			val = SPI_getvalue(tuple, tupdesc, i + 1);
@@ -1911,46 +1846,34 @@ get_sql_insert(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pka
 	}
 	appendStringInfo(&buf, ")");
 
-	relation_close(rel, AccessShareLock);
 	return (buf.data);
 }
 
 static char *
-get_sql_delete(Oid relid, int2vector *pkattnums, int16 pknumatts, char **tgt_pkattvals)
+get_sql_delete(Relation rel, int *pkattnums, int pknumatts, char **tgt_pkattvals)
 {
-	Relation	rel;
 	char	   *relname;
 	TupleDesc	tupdesc;
-	int			natts;
 	StringInfoData buf;
 	int			i;
 
 	initStringInfo(&buf);
 
 	/* get relation name including any needed schema prefix and quoting */
-	relname = generate_relation_name(relid);
+	relname = generate_relation_name(rel);
 
-	/*
-	 * Open relation using relid
-	 */
-	rel = relation_open(relid, AccessShareLock);
 	tupdesc = rel->rd_att;
-	natts = tupdesc->natts;
 
 	appendStringInfo(&buf, "DELETE FROM %s WHERE ", relname);
 	for (i = 0; i < pknumatts; i++)
 	{
-		int16		pkattnum = pkattnums->values[i];
+		int			pkattnum = pkattnums[i];
 
 		if (i > 0)
 			appendStringInfo(&buf, " AND ");
 
 		appendStringInfoString(&buf,
-		   quote_ident_cstr(NameStr(tupdesc->attrs[pkattnum - 1]->attname)));
-
-		if (tgt_pkattvals == NULL)
-			/* internal error */
-			elog(ERROR, "target key array must not be NULL");
+		   quote_ident_cstr(NameStr(tupdesc->attrs[pkattnum]->attname)));
 
 		if (tgt_pkattvals[i] != NULL)
 			appendStringInfo(&buf, " = %s",
@@ -1959,37 +1882,31 @@ get_sql_delete(Oid relid, int2vector *pkattnums, int16 pknumatts, char **tgt_pka
 			appendStringInfo(&buf, " IS NULL");
 	}
 
-	relation_close(rel, AccessShareLock);
 	return (buf.data);
 }
 
 static char *
-get_sql_update(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pkattvals, char **tgt_pkattvals)
+get_sql_update(Relation rel, int *pkattnums, int pknumatts, char **src_pkattvals, char **tgt_pkattvals)
 {
-	Relation	rel;
 	char	   *relname;
 	HeapTuple	tuple;
 	TupleDesc	tupdesc;
 	int			natts;
 	StringInfoData buf;
 	char	   *val;
-	int16		key;
+	int			key;
 	int			i;
 	bool		needComma;
 
 	initStringInfo(&buf);
 
 	/* get relation name including any needed schema prefix and quoting */
-	relname = generate_relation_name(relid);
+	relname = generate_relation_name(rel);
 
-	/*
-	 * Open relation using relid
-	 */
-	rel = relation_open(relid, AccessShareLock);
 	tupdesc = rel->rd_att;
 	natts = tupdesc->natts;
 
-	tuple = get_tuple_of_interest(relid, pkattnums, pknumatts, src_pkattvals);
+	tuple = get_tuple_of_interest(rel, pkattnums, pknumatts, src_pkattvals);
 	if (!tuple)
 		ereport(ERROR,
 				(errcode(ERRCODE_CARDINALITY_VIOLATION),
@@ -1997,6 +1914,9 @@ get_sql_update(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pka
 
 	appendStringInfo(&buf, "UPDATE %s SET ", relname);
 
+	/*
+	 * Note: i is physical column number (counting from 0).
+	 */
 	needComma = false;
 	for (i = 0; i < natts; i++)
 	{
@@ -2009,12 +1929,9 @@ get_sql_update(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pka
 		appendStringInfo(&buf, "%s = ",
 					  quote_ident_cstr(NameStr(tupdesc->attrs[i]->attname)));
 
-		if (tgt_pkattvals != NULL)
-			key = get_attnum_pk_pos(pkattnums, pknumatts, i + 1);
-		else
-			key = -1;
+		key = get_attnum_pk_pos(pkattnums, pknumatts, i);
 
-		if (key > -1)
+		if (key >= 0)
 			val = tgt_pkattvals[key] ? pstrdup(tgt_pkattvals[key]) : NULL;
 		else
 			val = SPI_getvalue(tuple, tupdesc, i + 1);
@@ -2033,29 +1950,22 @@ get_sql_update(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pka
 
 	for (i = 0; i < pknumatts; i++)
 	{
-		int16		pkattnum = pkattnums->values[i];
+		int			pkattnum = pkattnums[i];
 
 		if (i > 0)
 			appendStringInfo(&buf, " AND ");
 
 		appendStringInfo(&buf, "%s",
-		   quote_ident_cstr(NameStr(tupdesc->attrs[pkattnum - 1]->attname)));
+		   quote_ident_cstr(NameStr(tupdesc->attrs[pkattnum]->attname)));
 
-		if (tgt_pkattvals != NULL)
-			val = tgt_pkattvals[i] ? pstrdup(tgt_pkattvals[i]) : NULL;
-		else
-			val = SPI_getvalue(tuple, tupdesc, pkattnum);
+		val = tgt_pkattvals[i];
 
 		if (val != NULL)
-		{
 			appendStringInfo(&buf, " = %s", quote_literal_cstr(val));
-			pfree(val);
-		}
 		else
 			appendStringInfo(&buf, " IS NULL");
 	}
 
-	relation_close(rel, AccessShareLock);
 	return (buf.data);
 }
 
@@ -2095,8 +2005,8 @@ quote_ident_cstr(char *rawstr)
 	return result;
 }
 
-static int16
-get_attnum_pk_pos(int2vector *pkattnums, int16 pknumatts, int16 key)
+static int
+get_attnum_pk_pos(int *pkattnums, int pknumatts, int key)
 {
 	int			i;
 
@@ -2104,35 +2014,23 @@ get_attnum_pk_pos(int2vector *pkattnums, int16 pknumatts, int16 key)
 	 * Not likely a long list anyway, so just scan for the value
 	 */
 	for (i = 0; i < pknumatts; i++)
-		if (key == pkattnums->values[i])
+		if (key == pkattnums[i])
 			return i;
 
 	return -1;
 }
 
 static HeapTuple
-get_tuple_of_interest(Oid relid, int2vector *pkattnums, int16 pknumatts, char **src_pkattvals)
+get_tuple_of_interest(Relation rel, int *pkattnums, int pknumatts, char **src_pkattvals)
 {
-	Relation	rel;
 	char	   *relname;
 	TupleDesc	tupdesc;
+	int			natts;
 	StringInfoData buf;
 	int			ret;
 	HeapTuple	tuple;
 	int			i;
 
-	initStringInfo(&buf);
-
-	/* get relation name including any needed schema prefix and quoting */
-	relname = generate_relation_name(relid);
-
-	/*
-	 * Open relation using relid
-	 */
-	rel = relation_open(relid, AccessShareLock);
-	tupdesc = CreateTupleDescCopy(rel->rd_att);
-	relation_close(rel, AccessShareLock);
-
 	/*
 	 * Connect to SPI manager
 	 */
@@ -2140,21 +2038,46 @@ get_tuple_of_interest(Oid relid, int2vector *pkattnums, int16 pknumatts, char **
 		/* internal error */
 		elog(ERROR, "SPI connect failure - returned %d", ret);
 
+	initStringInfo(&buf);
+
+	/* get relation name including any needed schema prefix and quoting */
+	relname = generate_relation_name(rel);
+
+	tupdesc = rel->rd_att;
+	natts = tupdesc->natts;
+
 	/*
-	 * Build sql statement to look up tuple of interest Use src_pkattvals as
-	 * the criteria.
+	 * Build sql statement to look up tuple of interest, ie, the one matching
+	 * src_pkattvals.  We used to use "SELECT *" here, but it's simpler to
+	 * generate a result tuple that matches the table's physical structure,
+	 * with NULLs for any dropped columns.  Otherwise we have to deal with
+	 * two different tupdescs and everything's very confusing.
 	 */
-	appendStringInfo(&buf, "SELECT * FROM %s WHERE ", relname);
+	appendStringInfoString(&buf, "SELECT ");
+
+	for (i = 0; i < natts; i++)
+	{
+		if (i > 0)
+			appendStringInfoString(&buf, ", ");
+
+		if (tupdesc->attrs[i]->attisdropped)
+			appendStringInfoString(&buf, "NULL");
+		else
+			appendStringInfoString(&buf,
+								   quote_ident_cstr(NameStr(tupdesc->attrs[i]->attname)));
+	}
+
+	appendStringInfo(&buf, " FROM %s WHERE ", relname);
 
 	for (i = 0; i < pknumatts; i++)
 	{
-		int16		pkattnum = pkattnums->values[i];
+		int			pkattnum = pkattnums[i];
 
 		if (i > 0)
 			appendStringInfo(&buf, " AND ");
 
 		appendStringInfoString(&buf,
-		   quote_ident_cstr(NameStr(tupdesc->attrs[pkattnum - 1]->attname)));
+		   quote_ident_cstr(NameStr(tupdesc->attrs[pkattnum]->attname)));
 
 		if (src_pkattvals[i] != NULL)
 			appendStringInfo(&buf, " = %s",
@@ -2202,52 +2125,49 @@ get_tuple_of_interest(Oid relid, int2vector *pkattnums, int16 pknumatts, char **
 	return NULL;
 }
 
-static Oid
-get_relid_from_relname(text *relname_text)
+/*
+ * Open the relation named by relname_text, acquire specified type of lock,
+ * verify we have specified permissions.
+ * Caller must close rel when done with it.
+ */
+static Relation
+get_rel_from_relname(text *relname_text, LOCKMODE lockmode, AclMode aclmode)
 {
 	RangeVar   *relvar;
 	Relation	rel;
-	Oid			relid;
+	AclResult	aclresult;
 
 	relvar = makeRangeVarFromNameList(textToQualifiedNameList(relname_text));
-	rel = heap_openrv(relvar, AccessShareLock);
-	relid = RelationGetRelid(rel);
-	relation_close(rel, AccessShareLock);
+	rel = heap_openrv(relvar, lockmode);
 
-	return relid;
+	aclresult = pg_class_aclcheck(RelationGetRelid(rel), GetUserId(),
+								  aclmode);
+	if (aclresult != ACLCHECK_OK)
+		aclcheck_error(aclresult, ACL_KIND_CLASS,
+					   RelationGetRelationName(rel));
+
+	return rel;
 }
 
 /*
  * generate_relation_name - copied from ruleutils.c
- *		Compute the name to display for a relation specified by OID
+ *		Compute the name to display for a relation
  *
  * The result includes all necessary quoting and schema-prefixing.
  */
 static char *
-generate_relation_name(Oid relid)
+generate_relation_name(Relation rel)
 {
-	HeapTuple	tp;
-	Form_pg_class reltup;
 	char	   *nspname;
 	char	   *result;
 
-	tp = SearchSysCache(RELOID,
-						ObjectIdGetDatum(relid),
-						0, 0, 0);
-	if (!HeapTupleIsValid(tp))
-		elog(ERROR, "cache lookup failed for relation %u", relid);
-
-	reltup = (Form_pg_class) GETSTRUCT(tp);
-
 	/* Qualify the name if not visible in search path */
-	if (RelationIsVisible(relid))
+	if (RelationIsVisible(RelationGetRelid(rel)))
 		nspname = NULL;
 	else
-		nspname = get_namespace_name(reltup->relnamespace);
+		nspname = get_namespace_name(rel->rd_rel->relnamespace);
 
-	result = quote_qualified_identifier(nspname, NameStr(reltup->relname));
-
-	ReleaseSysCache(tp);
+	result = quote_qualified_identifier(nspname, RelationGetRelationName(rel));
 
 	return result;
 }
@@ -2257,13 +2177,13 @@ static remoteConn *
 getConnectionByName(const char *name)
 {
 	remoteConnHashEnt *hentry;
-	char		key[NAMEDATALEN];
+	char	   *key;
 
 	if (!remoteConnHash)
 		remoteConnHash = createConnHash();
 
-	MemSet(key, 0, NAMEDATALEN);
-	snprintf(key, NAMEDATALEN - 1, "%s", name);
+	key = pstrdup(name);
+	truncate_identifier(key, strlen(key), false);
 	hentry = (remoteConnHashEnt *) hash_search(remoteConnHash,
 											   key, HASH_FIND, NULL);
 
@@ -2289,20 +2209,25 @@ createNewConnection(const char *name, remoteConn * rconn)
 {
 	remoteConnHashEnt *hentry;
 	bool		found;
-	char		key[NAMEDATALEN];
+	char	   *key;
 
 	if (!remoteConnHash)
 		remoteConnHash = createConnHash();
 
-	MemSet(key, 0, NAMEDATALEN);
-	snprintf(key, NAMEDATALEN - 1, "%s", name);
+	key = pstrdup(name);
+	truncate_identifier(key, strlen(key), true);
 	hentry = (remoteConnHashEnt *) hash_search(remoteConnHash, key,
 											   HASH_ENTER, &found);
 
 	if (found)
+	{
+		PQfinish(rconn->conn);
+		pfree(rconn);
+
 		ereport(ERROR,
 				(errcode(ERRCODE_DUPLICATE_OBJECT),
 				 errmsg("duplicate connection name")));
+	}
 
 	hentry->rconn = rconn;
 	strlcpy(hentry->name, name, sizeof(hentry->name));
@@ -2313,14 +2238,13 @@ deleteConnection(const char *name)
 {
 	remoteConnHashEnt *hentry;
 	bool		found;
-	char		key[NAMEDATALEN];
+	char	   *key;
 
 	if (!remoteConnHash)
 		remoteConnHash = createConnHash();
 
-	MemSet(key, 0, NAMEDATALEN);
-	snprintf(key, NAMEDATALEN - 1, "%s", name);
-
+	key = pstrdup(name);
+	truncate_identifier(key, strlen(key), false);
 	hentry = (remoteConnHashEnt *) hash_search(remoteConnHash,
 											   key, HASH_REMOVE, &found);
 
@@ -2351,27 +2275,52 @@ dblink_security_check(PGconn *conn, remoteConn *rconn)
 	}
 }
 
-static int
-get_nondropped_natts(Oid relid)
+/*
+ * Validate the PK-attnums argument for dblink_build_sql_insert() and related
+ * functions, and translate to the internal representation.
+ *
+ * The user supplies an int2vector of 1-based physical attnums, plus a count
+ * argument (the need for the separate count argument is historical, but we
+ * still check it).  We check that each attnum corresponds to a valid,
+ * non-dropped attribute of the rel.  We do *not* prevent attnums from being
+ * listed twice, though the actual use-case for such things is dubious.
+ *
+ * The internal representation is a palloc'd int array of 0-based physical
+ * attnums.
+ */
+static void
+validate_pkattnums(Relation rel,
+				   int2vector *pkattnums_arg, int32 pknumatts_arg,
+				   int **pkattnums, int *pknumatts)
 {
-	int			nondropped_natts = 0;
-	TupleDesc	tupdesc;
-	Relation	rel;
-	int			natts;
+	TupleDesc	tupdesc = rel->rd_att;
+	int			natts = tupdesc->natts;
 	int			i;
 
-	rel = relation_open(relid, AccessShareLock);
-	tupdesc = rel->rd_att;
-	natts = tupdesc->natts;
+	/* Don't take more array elements than there are */
+	pknumatts_arg = Min(pknumatts_arg, pkattnums_arg->dim1);
 
-	for (i = 0; i < natts; i++)
+	/* Must have at least one pk attnum selected */
+	if (pknumatts_arg <= 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("number of key attributes must be > 0")));
+
+	/* Allocate output array */
+	*pkattnums = (int *) palloc(pknumatts_arg * sizeof(int));
+	*pknumatts = pknumatts_arg;
+
+	/* Validate attnums and convert to internal form */
+	for (i = 0; i < pknumatts_arg; i++)
 	{
-		if (tupdesc->attrs[i]->attisdropped)
-			continue;
-		nondropped_natts++;
-	}
+		int		pkattnum = pkattnums_arg->values[i];
 
-	relation_close(rel, AccessShareLock);
-	return nondropped_natts;
+		if (pkattnum <= 0 || pkattnum > natts ||
+			tupdesc->attrs[pkattnum - 1]->attisdropped)
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("invalid attribute number %d", pkattnum)));
+		(*pkattnums)[i] = pkattnum - 1;
+	}
 }
 
