@@ -3,12 +3,12 @@
  * hio.c
  *	  POSTGRES heap access method input/output code.
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Id: hio.c,v 1.46.2.1 2005/05/07 21:33:47 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/access/heap/hio.c,v 1.68 2008/01/01 19:45:46 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -23,7 +23,7 @@
 /*
  * RelationPutHeapTuple - place tuple at specified page
  *
- * !!! ELOG(ERROR) IS DISALLOWED HERE !!!
+ * !!! EREPORT(ERROR) IS DISALLOWED HERE !!!  Must PANIC on failure!!!
  *
  * Note - caller must hold BUFFER_LOCK_EXCLUSIVE on the buffer.
  */
@@ -37,20 +37,14 @@ RelationPutHeapTuple(Relation relation,
 	ItemId		itemId;
 	Item		item;
 
-	/*
-	 * increment access statistics
-	 */
-	IncrHeapAccessStat(local_RelationPutHeapTuple);
-	IncrHeapAccessStat(global_RelationPutHeapTuple);
-
 	/* Add the tuple to the page */
 	pageHeader = BufferGetPage(buffer);
 
 	offnum = PageAddItem(pageHeader, (Item) tuple->t_data,
-						 tuple->t_len, InvalidOffsetNumber, LP_USED);
+						 tuple->t_len, InvalidOffsetNumber, false, true);
 
 	if (offnum == InvalidOffsetNumber)
-		elog(PANIC, "RelationPutHeapTuple: failed to add tuple");
+		elog(PANIC, "failed to add tuple to page");
 
 	/* Update tuple->t_self to the actual position where it was stored */
 	ItemPointerSet(&(tuple->t_self), BufferGetBlockNumber(buffer), offnum);
@@ -85,21 +79,36 @@ RelationPutHeapTuple(Relation relation,
  *	happen if space is freed in that page after heap_update finds there's not
  *	enough there).	In that case, the page will be pinned and locked only once.
  *
- *	Note that we use LockPage(rel, 0) to lock relation for extension.
- *	We can do this as long as in all other places we use page-level locking
- *	for indices only. Alternatively, we could define pseudo-table as
- *	we do for transactions with XactLockTable.
+ *	If use_fsm is true (the normal case), we use FSM to help us find free
+ *	space.	If use_fsm is false, we always append a new empty page to the
+ *	end of the relation if the tuple won't fit on the current target page.
+ *	This can save some cycles when we know the relation is new and doesn't
+ *	contain useful amounts of free space.
  *
- *	ELOG(ERROR) is allowed here, so this routine *must* be called
+ *	The use_fsm = false case is also useful for non-WAL-logged additions to a
+ *	relation, if the caller holds exclusive lock and is careful to invalidate
+ *	relation->rd_targblock before the first insertion --- that ensures that
+ *	all insertions will occur into newly added pages and not be intermixed
+ *	with tuples from other transactions.  That way, a crash can't risk losing
+ *	any committed data of other transactions.  (See heap_insert's comments
+ *	for additional constraints needed for safe usage of this behavior.)
+ *
+ *	We always try to avoid filling existing pages further than the fillfactor.
+ *	This is OK since this routine is not consulted when updating a tuple and
+ *	keeping it on the same page, which is the scenario fillfactor is meant
+ *	to reserve space for.
+ *
+ *	ereport(ERROR) is allowed here, so this routine *must* be called
  *	before any (unlogged) changes are made in buffer pool.
  */
 Buffer
 RelationGetBufferForTuple(Relation relation, Size len,
-						  Buffer otherBuffer)
+						  Buffer otherBuffer, bool use_fsm)
 {
 	Buffer		buffer = InvalidBuffer;
 	Page		pageHeader;
-	Size		pageFreeSpace;
+	Size		pageFreeSpace,
+				saveFreeSpace;
 	BlockNumber targetBlock,
 				otherBlock;
 	bool		needLock;
@@ -109,42 +118,57 @@ RelationGetBufferForTuple(Relation relation, Size len,
 	/*
 	 * If we're gonna fail for oversize tuple, do it right away
 	 */
-	if (len > MaxTupleSize)
-		elog(ERROR, "Tuple is too big: size %lu, max size %ld",
-			 (unsigned long) len, MaxTupleSize);
+	if (len > MaxHeapTupleSize)
+		ereport(ERROR,
+				(errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
+				 errmsg("row is too big: size %lu, maximum size %lu",
+						(unsigned long) len,
+						(unsigned long) MaxHeapTupleSize)));
+
+	/* Compute desired extra freespace due to fillfactor option */
+	saveFreeSpace = RelationGetTargetPageFreeSpace(relation,
+												   HEAP_DEFAULT_FILLFACTOR);
 
 	if (otherBuffer != InvalidBuffer)
 		otherBlock = BufferGetBlockNumber(otherBuffer);
 	else
-		otherBlock = InvalidBlockNumber;		/* just to keep compiler
-												 * quiet */
+		otherBlock = InvalidBlockNumber;		/* just to keep compiler quiet */
 
 	/*
-	 * We first try to put the tuple on the same page we last inserted a
-	 * tuple on, as cached in the relcache entry.  If that doesn't work,
-	 * we ask the shared Free Space Map to locate a suitable page.	Since
-	 * the FSM's info might be out of date, we have to be prepared to loop
-	 * around and retry multiple times.  (To insure this isn't an infinite
-	 * loop, we must update the FSM with the correct amount of free space
-	 * on each page that proves not to be suitable.)  If the FSM has no
-	 * record of a page with enough free space, we give up and extend the
-	 * relation.
+	 * We first try to put the tuple on the same page we last inserted a tuple
+	 * on, as cached in the relcache entry.  If that doesn't work, we ask the
+	 * shared Free Space Map to locate a suitable page.  Since the FSM's info
+	 * might be out of date, we have to be prepared to loop around and retry
+	 * multiple times.	(To insure this isn't an infinite loop, we must update
+	 * the FSM with the correct amount of free space on each page that proves
+	 * not to be suitable.)  If the FSM has no record of a page with enough
+	 * free space, we give up and extend the relation.
+	 *
+	 * When use_fsm is false, we either put the tuple onto the existing target
+	 * page or extend the relation.
 	 */
+	if (len + saveFreeSpace <= MaxHeapTupleSize)
+		targetBlock = relation->rd_targblock;
+	else
+	{
+		/* can't fit, don't screw up FSM request tracking by trying */
+		targetBlock = InvalidBlockNumber;
+		use_fsm = false;
+	}
 
-	targetBlock = relation->rd_targblock;
-
-	if (targetBlock == InvalidBlockNumber)
+	if (targetBlock == InvalidBlockNumber && use_fsm)
 	{
 		/*
 		 * We have no cached target page, so ask the FSM for an initial
 		 * target.
 		 */
-		targetBlock = GetPageWithFreeSpace(&relation->rd_node, len);
+		targetBlock = GetPageWithFreeSpace(&relation->rd_node,
+										   len + saveFreeSpace);
 
 		/*
-		 * If the FSM knows nothing of the rel, try the last page before
-		 * we give up and extend.  This avoids one-tuple-per-page syndrome
-		 * during bootstrapping or in a recently-started system.
+		 * If the FSM knows nothing of the rel, try the last page before we
+		 * give up and extend.	This avoids one-tuple-per-page syndrome during
+		 * bootstrapping or in a recently-started system.
 		 */
 		if (targetBlock == InvalidBlockNumber)
 		{
@@ -159,8 +183,8 @@ RelationGetBufferForTuple(Relation relation, Size len,
 	{
 		/*
 		 * Read and exclusive-lock the target block, as well as the other
-		 * block if one was given, taking suitable care with lock ordering
-		 * and the possibility they are the same block.
+		 * block if one was given, taking suitable care with lock ordering and
+		 * the possibility they are the same block.
 		 */
 		if (otherBuffer == InvalidBuffer)
 		{
@@ -190,12 +214,12 @@ RelationGetBufferForTuple(Relation relation, Size len,
 		}
 
 		/*
-		 * Now we can check to see if there's enough free space here. If
-		 * so, we're done.
+		 * Now we can check to see if there's enough free space here. If so,
+		 * we're done.
 		 */
 		pageHeader = (Page) BufferGetPage(buffer);
-		pageFreeSpace = PageGetFreeSpace(pageHeader);
-		if (len <= pageFreeSpace)
+		pageFreeSpace = PageGetHeapFreeSpace(pageHeader);
+		if (len + saveFreeSpace <= pageFreeSpace)
 		{
 			/* use this page as future insert target, too */
 			relation->rd_targblock = targetBlock;
@@ -204,9 +228,9 @@ RelationGetBufferForTuple(Relation relation, Size len,
 
 		/*
 		 * Not enough space, so we must give up our page locks and pin (if
-		 * any) and prepare to look elsewhere.	We don't care which order
-		 * we unlock the two buffers in, so this can be slightly simpler
-		 * than the code above.
+		 * any) and prepare to look elsewhere.	We don't care which order we
+		 * unlock the two buffers in, so this can be slightly simpler than the
+		 * code above.
 		 */
 		LockBuffer(buffer, BUFFER_LOCK_UNLOCK);
 		if (otherBuffer == InvalidBuffer)
@@ -217,41 +241,44 @@ RelationGetBufferForTuple(Relation relation, Size len,
 			ReleaseBuffer(buffer);
 		}
 
+		/* Without FSM, always fall out of the loop and extend */
+		if (!use_fsm)
+			break;
+
 		/*
-		 * Update FSM as to condition of this page, and ask for another
-		 * page to try.
+		 * Update FSM as to condition of this page, and ask for another page
+		 * to try.
 		 */
 		targetBlock = RecordAndGetPageWithFreeSpace(&relation->rd_node,
 													targetBlock,
 													pageFreeSpace,
-													len);
+													len + saveFreeSpace);
 	}
 
 	/*
 	 * Have to extend the relation.
 	 *
-	 * We have to use a lock to ensure no one else is extending the rel at
-	 * the same time, else we will both try to initialize the same new
-	 * page.  We can skip locking for new or temp relations, however,
-	 * since no one else could be accessing them.
+	 * We have to use a lock to ensure no one else is extending the rel at the
+	 * same time, else we will both try to initialize the same new page.  We
+	 * can skip locking for new or temp relations, however, since no one else
+	 * could be accessing them.
 	 */
-	needLock = !(relation->rd_isnew || relation->rd_istemp);
+	needLock = !RELATION_IS_LOCAL(relation);
 
 	if (needLock)
-		LockPage(relation, 0, ExclusiveLock);
+		LockRelationForExtension(relation, ExclusiveLock);
 
 	/*
-	 * XXX This does an lseek - rather expensive - but at the moment it is
-	 * the only way to accurately determine how many blocks are in a
-	 * relation.  Is it worth keeping an accurate file length in shared
-	 * memory someplace, rather than relying on the kernel to do it for
-	 * us?
+	 * XXX This does an lseek - rather expensive - but at the moment it is the
+	 * only way to accurately determine how many blocks are in a relation.	Is
+	 * it worth keeping an accurate file length in shared memory someplace,
+	 * rather than relying on the kernel to do it for us?
 	 */
 	buffer = ReadBuffer(relation, P_NEW);
 
 	/*
-	 * We can be certain that locking the otherBuffer first is OK, since
-	 * it must have a lower page number.
+	 * We can be certain that locking the otherBuffer first is OK, since it
+	 * must have a lower page number.
 	 */
 	if (otherBuffer != InvalidBuffer)
 		LockBuffer(otherBuffer, BUFFER_LOCK_EXCLUSIVE);
@@ -262,25 +289,32 @@ RelationGetBufferForTuple(Relation relation, Size len,
 	LockBuffer(buffer, BUFFER_LOCK_EXCLUSIVE);
 
 	/*
-	 * Release the file-extension lock; it's now OK for someone else to
-	 * extend the relation some more.  Note that we cannot release this
-	 * lock before we have buffer lock on the new page, or we risk a
-	 * race condition against vacuumlazy.c --- see comments therein.
+	 * Release the file-extension lock; it's now OK for someone else to extend
+	 * the relation some more.	Note that we cannot release this lock before
+	 * we have buffer lock on the new page, or we risk a race condition
+	 * against vacuumlazy.c --- see comments therein.
 	 */
 	if (needLock)
-		UnlockPage(relation, 0, ExclusiveLock);
+		UnlockRelationForExtension(relation, ExclusiveLock);
 
 	/*
-	 * We need to initialize the empty new page.
+	 * We need to initialize the empty new page.  Double-check that it really
+	 * is empty (this should never happen, but if it does we don't want to
+	 * risk wiping out valid data).
 	 */
 	pageHeader = (Page) BufferGetPage(buffer);
-	Assert(PageIsNew((PageHeader) pageHeader));
+
+	if (!PageIsNew((PageHeader) pageHeader))
+		elog(ERROR, "page %u of relation \"%s\" should be empty but is not",
+			 BufferGetBlockNumber(buffer),
+			 RelationGetRelationName(relation));
+
 	PageInit(pageHeader, BufferGetPageSize(buffer), 0);
 
-	if (len > PageGetFreeSpace(pageHeader))
+	if (len > PageGetHeapFreeSpace(pageHeader))
 	{
 		/* We should not get here given the test at the top */
-		elog(PANIC, "Tuple is too big: size %lu", (unsigned long) len);
+		elog(PANIC, "tuple is too big: size %lu", (unsigned long) len);
 	}
 
 	/*
