@@ -83,12 +83,12 @@
  * bootparse.y
  *	  yacc grammar for the "bootstrap" mode (BKI file format)
  *
- * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/bootstrap/bootparse.y,v 1.84 2006/08/25 04:06:46 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/bootstrap/bootparse.y,v 1.91 2008/01/01 19:45:48 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -576,9 +576,9 @@ static const yytype_uint16 yyrline[] =
 {
        0,   114,   114,   115,   119,   120,   124,   125,   126,   127,
      128,   129,   130,   131,   135,   144,   150,   160,   170,   159,
-     224,   223,   248,   266,   284,   294,   304,   305,   309,   320,
-     321,   325,   326,   330,   331,   335,   336,   340,   349,   353,
-     354,   358,   359,   360,   364,   366,   368,   373,   377
+     224,   223,   245,   263,   281,   291,   301,   302,   306,   319,
+     320,   324,   325,   329,   330,   334,   335,   339,   348,   352,
+     353,   357,   358,   359,   363,   365,   367,   372,   376
 };
 #endif
 
@@ -1670,10 +1670,7 @@ yyreduce:
 						elog(ERROR, "incorrect number of columns in row (expected %d, got %d)",
 							 numattr, num_columns_read);
 					if (boot_reldesc == NULL)
-					{
-						elog(ERROR, "relation not open");
-						err_out();
-					}
+						elog(FATAL, "relation not open");
 					InsertOneTuple((yyvsp[(2) - (6)].oidval));
 					do_end();
 				}
@@ -1682,7 +1679,7 @@ yyreduce:
   case 22:
 
 /* Line 1455 of yacc.c  */
-#line 249 "bootparse.y"
+#line 246 "bootparse.y"
     {
 					do_start();
 
@@ -1692,7 +1689,7 @@ yyreduce:
 								LexIDStr((yyvsp[(8) - (11)].ival)),
 								NULL,
 								(yyvsp[(10) - (11)].list),
-								NULL, NIL, NIL,
+								NULL, NIL,
 								false, false, false,
 								false, false, true, false, false);
 					do_end();
@@ -1702,7 +1699,7 @@ yyreduce:
   case 23:
 
 /* Line 1455 of yacc.c  */
-#line 267 "bootparse.y"
+#line 264 "bootparse.y"
     {
 					do_start();
 
@@ -1712,7 +1709,7 @@ yyreduce:
 								LexIDStr((yyvsp[(9) - (12)].ival)),
 								NULL,
 								(yyvsp[(11) - (12)].list),
-								NULL, NIL, NIL,
+								NULL, NIL,
 								true, false, false,
 								false, false, true, false, false);
 					do_end();
@@ -1722,7 +1719,7 @@ yyreduce:
   case 24:
 
 /* Line 1455 of yacc.c  */
-#line 285 "bootparse.y"
+#line 282 "bootparse.y"
     {
 					do_start();
 
@@ -1734,7 +1731,7 @@ yyreduce:
   case 25:
 
 /* Line 1455 of yacc.c  */
-#line 295 "bootparse.y"
+#line 292 "bootparse.y"
     {
 					do_start();
 					build_indices();
@@ -1745,26 +1742,28 @@ yyreduce:
   case 26:
 
 /* Line 1455 of yacc.c  */
-#line 304 "bootparse.y"
+#line 301 "bootparse.y"
     { (yyval.list) = lappend((yyvsp[(1) - (3)].list), (yyvsp[(3) - (3)].ielem)); }
     break;
 
   case 27:
 
 /* Line 1455 of yacc.c  */
-#line 305 "bootparse.y"
+#line 302 "bootparse.y"
     { (yyval.list) = list_make1((yyvsp[(1) - (1)].ielem)); }
     break;
 
   case 28:
 
 /* Line 1455 of yacc.c  */
-#line 310 "bootparse.y"
+#line 307 "bootparse.y"
     {
 					IndexElem *n = makeNode(IndexElem);
 					n->name = LexIDStr((yyvsp[(1) - (2)].ival));
 					n->expr = NULL;
 					n->opclass = list_make1(makeString(LexIDStr((yyvsp[(2) - (2)].ival))));
+					n->ordering = SORTBY_DEFAULT;
+					n->nulls_ordering = SORTBY_NULLS_DEFAULT;
 					(yyval.ielem) = n;
 				}
     break;
@@ -1772,49 +1771,49 @@ yyreduce:
   case 29:
 
 /* Line 1455 of yacc.c  */
-#line 320 "bootparse.y"
+#line 319 "bootparse.y"
     { (yyval.ival) = 1; }
     break;
 
   case 30:
 
 /* Line 1455 of yacc.c  */
-#line 321 "bootparse.y"
+#line 320 "bootparse.y"
     { (yyval.ival) = 0; }
     break;
 
   case 31:
 
 /* Line 1455 of yacc.c  */
-#line 325 "bootparse.y"
+#line 324 "bootparse.y"
     { (yyval.ival) = 1; }
     break;
 
   case 32:
 
 /* Line 1455 of yacc.c  */
-#line 326 "bootparse.y"
+#line 325 "bootparse.y"
     { (yyval.ival) = 0; }
     break;
 
   case 33:
 
 /* Line 1455 of yacc.c  */
-#line 330 "bootparse.y"
+#line 329 "bootparse.y"
     { (yyval.ival) = 1; }
     break;
 
   case 34:
 
 /* Line 1455 of yacc.c  */
-#line 331 "bootparse.y"
+#line 330 "bootparse.y"
     { (yyval.ival) = 0; }
     break;
 
   case 37:
 
 /* Line 1455 of yacc.c  */
-#line 341 "bootparse.y"
+#line 340 "bootparse.y"
     {
 				   if (++numattr > MAXATTR)
 						elog(FATAL, "too many columns");
@@ -1825,63 +1824,63 @@ yyreduce:
   case 38:
 
 /* Line 1455 of yacc.c  */
-#line 349 "bootparse.y"
+#line 348 "bootparse.y"
     { (yyval.oidval) = atooid(LexIDStr((yyvsp[(1) - (1)].ival))); }
     break;
 
   case 39:
 
 /* Line 1455 of yacc.c  */
-#line 353 "bootparse.y"
+#line 352 "bootparse.y"
     { (yyval.oidval) = (yyvsp[(3) - (3)].oidval); }
     break;
 
   case 40:
 
 /* Line 1455 of yacc.c  */
-#line 354 "bootparse.y"
+#line 353 "bootparse.y"
     { (yyval.oidval) = (Oid) 0; }
     break;
 
   case 44:
 
 /* Line 1455 of yacc.c  */
-#line 365 "bootparse.y"
+#line 364 "bootparse.y"
     { InsertOneValue(LexIDStr((yyvsp[(1) - (1)].ival)), num_columns_read++); }
     break;
 
   case 45:
 
 /* Line 1455 of yacc.c  */
-#line 367 "bootparse.y"
+#line 366 "bootparse.y"
     { InsertOneValue(LexIDStr((yyvsp[(1) - (1)].ival)), num_columns_read++); }
     break;
 
   case 46:
 
 /* Line 1455 of yacc.c  */
-#line 369 "bootparse.y"
+#line 368 "bootparse.y"
     { InsertOneNull(num_columns_read++); }
     break;
 
   case 47:
 
 /* Line 1455 of yacc.c  */
-#line 373 "bootparse.y"
+#line 372 "bootparse.y"
     { (yyval.ival)=yylval.ival; }
     break;
 
   case 48:
 
 /* Line 1455 of yacc.c  */
-#line 377 "bootparse.y"
+#line 376 "bootparse.y"
     { (yyval.ival)=yylval.ival; }
     break;
 
 
 
 /* Line 1455 of yacc.c  */
-#line 1885 "y.tab.c"
+#line 1884 "y.tab.c"
       default: break;
     }
   YY_SYMBOL_PRINT ("-> $$ =", yyr1[yyn], &yyval, &yyloc);
@@ -2093,7 +2092,7 @@ yyreturn:
 
 
 /* Line 1675 of yacc.c  */
-#line 379 "bootparse.y"
+#line 378 "bootparse.y"
 
 
 #include "bootscanner.c"
