@@ -1427,6 +1427,7 @@ _outEquivalenceMember(StringInfo str, EquivalenceMember *node)
 
 	WRITE_NODE_FIELD(em_expr);
 	WRITE_BITMAPSET_FIELD(em_relids);
+	WRITE_BITMAPSET_FIELD(em_nullable_relids);
 	WRITE_BOOL_FIELD(em_is_const);
 	WRITE_BOOL_FIELD(em_is_child);
 	WRITE_OID_FIELD(em_datatype);
@@ -1629,6 +1630,15 @@ _outDefElem(StringInfo str, DefElem *node)
 	WRITE_NODE_FIELD(arg);
 }
 
+static void
+_outInhRelation(StringInfo str, InhRelation *node)
+{
+	WRITE_NODE_TYPE("INHRELATION");
+
+	WRITE_NODE_FIELD(relation);
+	WRITE_NODE_FIELD(options);
+}
+
 static void
 _outLockingClause(StringInfo str, LockingClause *node)
 {
@@ -2435,6 +2445,9 @@ _outNode(StringInfo str, void *obj)
 			case T_DefElem:
 				_outDefElem(str, obj);
 				break;
+			case T_InhRelation:
+				_outInhRelation(str, obj);
+				break;
 			case T_LockingClause:
 				_outLockingClause(str, obj);
 				break;
