@@ -14,23 +14,23 @@ struct ECPGstruct_member
 struct ECPGtype
 {
 	enum ECPGttype type;
-	char	   *size;			/* For array it is the number of elements.
-								 * For varchar it is the maxsize of the
-								 * area. */
-	char	   *struct_sizeof;	/* For a struct this is the sizeof() type
-								 * as string */
+	char	   *size;			/* For array it is the number of elements. For
+								 * varchar it is the maxsize of the area. */
+	char	   *struct_sizeof;	/* For a struct this is the sizeof() type as
+								 * string */
 	union
 	{
-		struct ECPGtype *element;		/* For an array this is the type
-										 * of the element */
+		struct ECPGtype *element;		/* For an array this is the type of
+										 * the element */
 		struct ECPGstruct_member *members;		/* A pointer to a list of
 												 * members. */
 	}			u;
+	int			lineno;
 };
 
 /* Everything is malloced. */
 void		ECPGmake_struct_member(char *, struct ECPGtype *, struct ECPGstruct_member **);
-struct ECPGtype *ECPGmake_simple_type(enum ECPGttype, char *);
+struct ECPGtype *ECPGmake_simple_type(enum ECPGttype, char *, int);
 struct ECPGtype *ECPGmake_varchar_type(enum ECPGttype, long);
 struct ECPGtype *ECPGmake_array_type(struct ECPGtype *, char *);
 struct ECPGtype *ECPGmake_struct_type(struct ECPGstruct_member *, enum ECPGttype, char *);
@@ -61,7 +61,7 @@ struct ECPGtemp_type
 	const char *name;
 };
 
-extern const char *ECPGtype_name(enum ECPGttype type);
+extern const char *ecpg_type_name(enum ECPGttype type);
 
 /* some stuff for whenever statements */
 enum WHEN_TYPE
@@ -95,6 +95,13 @@ struct su_symbol
 	char	   *symbol;
 };
 
+struct prep
+{
+	char	   *name;
+	char	   *stmt;
+	char	   *type;
+};
+
 struct this_type
 {
 	enum ECPGttype type_enum;
