@@ -3,12 +3,12 @@
  * path.c
  *	  portable path handling routines
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/port/path.c,v 1.61.2.2 2005/12/23 22:34:27 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/port/path.c,v 1.74.2.2 2009/04/03 23:27:22 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -69,7 +69,7 @@ skip_drive(const char *path)
 		while (*path && !IS_DIR_SEP(*path))
 			path++;
 	}
-	else if (isalpha(path[0]) && path[1] == ':')
+	else if (isalpha((unsigned char) path[0]) && path[1] == ':')
 	{
 		path += 2;
 	}
@@ -165,6 +165,8 @@ make_native_path(char *filename)
 /*
  * join_path_components - join two path components, inserting a slash
  *
+ * We omit the slash if either given component is empty.
+ *
  * ret_path is the output area (must be of size MAXPGPATH)
  *
  * ret_path can be the same as head, but not the same as tail.
@@ -174,40 +176,25 @@ join_path_components(char *ret_path,
 					 const char *head, const char *tail)
 {
 	if (ret_path != head)
-		StrNCpy(ret_path, head, MAXPGPATH);
+		strlcpy(ret_path, head, MAXPGPATH);
 
 	/*
-	 * Remove any leading "." and ".." in the tail component, adjusting head
-	 * as needed.
+	 * Remove any leading "." in the tail component.
+	 *
+	 * Note: we used to try to remove ".." as well, but that's tricky to get
+	 * right; now we just leave it to be done by canonicalize_path() later.
 	 */
-	for (;;)
-	{
-		if (tail[0] == '.' && IS_DIR_SEP(tail[1]))
-		{
-			tail += 2;
-		}
-		else if (tail[0] == '.' && tail[1] == '\0')
-		{
-			tail += 1;
-			break;
-		}
-		else if (tail[0] == '.' && tail[1] == '.' && IS_DIR_SEP(tail[2]))
-		{
-			trim_directory(ret_path);
-			tail += 3;
-		}
-		else if (tail[0] == '.' && tail[1] == '.' && tail[2] == '\0')
-		{
-			trim_directory(ret_path);
-			tail += 2;
-			break;
-		}
-		else
-			break;
-	}
+	while (tail[0] == '.' && IS_DIR_SEP(tail[1]))
+		tail += 2;
+
 	if (*tail)
+	{
+		/* only separate with slash if head wasn't empty */
 		snprintf(ret_path + strlen(ret_path), MAXPGPATH - strlen(ret_path),
-				 "/%s", tail);
+				 "%s%s",
+				 (*(skip_drive(head)) != '\0') ? "/" : "",
+				 tail);
+	}
 }
 
 
@@ -389,6 +376,7 @@ const char *
 get_progname(const char *argv0)
 {
 	const char *nodir_name;
+	char	   *progname;
 
 	nodir_name = last_dir_separator(argv0);
 	if (nodir_name)
@@ -396,38 +384,45 @@ get_progname(const char *argv0)
 	else
 		nodir_name = skip_drive(argv0);
 
-#if defined(__CYGWIN__) || defined(WIN32)
-	/* strip .exe suffix, regardless of case */
-	if (strlen(nodir_name) > sizeof(EXE) - 1 &&
-		pg_strcasecmp(nodir_name + strlen(nodir_name) - (sizeof(EXE) - 1), EXE) == 0)
+	/*
+	 * Make a copy in case argv[0] is modified by ps_status. Leaks memory, but
+	 * called only once.
+	 */
+	progname = strdup(nodir_name);
+	if (progname == NULL)
 	{
-		char	   *progname;
-
-		progname = strdup(nodir_name);	/* leaks memory, but called only once */
-		if (progname == NULL)
-		{
-			fprintf(stderr, "%s: out of memory\n", nodir_name);
-			exit(1);			/* This could exit the postmaster */
-		}
-		progname[strlen(progname) - (sizeof(EXE) - 1)] = '\0';
-		nodir_name = progname;
+		fprintf(stderr, "%s: out of memory\n", nodir_name);
+		exit(1);				/* This could exit the postmaster */
 	}
+
+#if defined(__CYGWIN__) || defined(WIN32)
+	/* strip ".exe" suffix, regardless of case */
+	if (strlen(progname) > sizeof(EXE) - 1 &&
+	pg_strcasecmp(progname + strlen(progname) - (sizeof(EXE) - 1), EXE) == 0)
+		progname[strlen(progname) - (sizeof(EXE) - 1)] = '\0';
 #endif
 
-	return nodir_name;
+	return progname;
 }
 
 
 /*
- * dir_strcmp: strcmp except any two DIR_SEP characters are considered equal
+ * dir_strcmp: strcmp except any two DIR_SEP characters are considered equal,
+ * and we honor filesystem case insensitivity if known
  */
 static int
 dir_strcmp(const char *s1, const char *s2)
 {
 	while (*s1 && *s2)
 	{
-		if (*s1 != *s2 &&
-			!(IS_DIR_SEP(*s1) && IS_DIR_SEP(*s2)))
+		if (
+#ifndef WIN32
+			*s1 != *s2
+#else
+			/* On windows, paths are case-insensitive */
+			pg_tolower((unsigned char) *s1) != pg_tolower((unsigned char) *s2)
+#endif
+			&& !(IS_DIR_SEP(*s1) && IS_DIR_SEP(*s2)))
 			return (int) *s1 - (int) *s2;
 		s1++, s2++;
 	}
@@ -492,7 +487,7 @@ make_relative_path(char *ret_path, const char *target_path,
 	 * Set up my_exec_path without the actual executable name, and
 	 * canonicalize to simplify comparison to bin_path.
 	 */
-	StrNCpy(ret_path, my_exec_path, MAXPGPATH);
+	strlcpy(ret_path, my_exec_path, MAXPGPATH);
 	trim_directory(ret_path);	/* remove my executable name */
 	canonicalize_path(ret_path);
 
@@ -501,7 +496,7 @@ make_relative_path(char *ret_path, const char *target_path,
 	 */
 	tail_start = (int) strlen(ret_path) - tail_len;
 	if (tail_start > 0 &&
-		IS_DIR_SEP(ret_path[tail_start-1]) &&
+		IS_DIR_SEP(ret_path[tail_start - 1]) &&
 		dir_strcmp(ret_path + tail_start, bin_path + prefix_len) == 0)
 	{
 		ret_path[tail_start] = '\0';
@@ -512,7 +507,7 @@ make_relative_path(char *ret_path, const char *target_path,
 	}
 
 no_match:
-	StrNCpy(ret_path, target_path, MAXPGPATH);
+	strlcpy(ret_path, target_path, MAXPGPATH);
 	canonicalize_path(ret_path);
 }
 
@@ -624,13 +619,18 @@ get_home_path(char *ret_path)
 
 	if (pqGetpwuid(geteuid(), &pwdstr, pwdbuf, sizeof(pwdbuf), &pwd) != 0)
 		return false;
-	StrNCpy(ret_path, pwd->pw_dir, MAXPGPATH);
+	strlcpy(ret_path, pwd->pw_dir, MAXPGPATH);
 	return true;
 #else
-	char		tmppath[MAX_PATH];
+	char	   *tmppath;
 
-	ZeroMemory(tmppath, sizeof(tmppath));
-	if (SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, 0, tmppath) != S_OK)
+	/*
+	 * Note: We use getenv here because the more modern
+	 * SHGetSpecialFolderPath() will force us to link with shell32.lib which
+	 * eats valuable desktop heap.
+	 */
+	tmppath = getenv("APPDATA");
+	if (!tmppath)
 		return false;
 	snprintf(ret_path, MAXPGPATH, "%s/postgresql", tmppath);
 	return true;
@@ -643,6 +643,15 @@ get_home_path(char *ret_path)
  *
  * Modify the given string in-place to name the parent directory of the
  * named file.
+ *
+ * If the input is just a file name with no directory part, the result is
+ * an empty string, not ".".  This is appropriate when the next step is
+ * join_path_components(), but might need special handling otherwise.
+ *
+ * Caution: this will not produce desirable results if the string ends
+ * with "..".  For most callers this is not a problem since the string
+ * is already known to name a regular file.  If in doubt, apply
+ * canonicalize_path() first.
  */
 void
 get_parent_directory(char *path)
@@ -651,54 +660,6 @@ get_parent_directory(char *path)
 }
 
 
-/*
- *	set_pglocale_pgservice
- *
- *	Set application-specific locale and service directory
- *
- *	This function takes an argv[0] rather than a full path.
- */
-void
-set_pglocale_pgservice(const char *argv0, const char *app)
-{
-	char		path[MAXPGPATH];
-	char		my_exec_path[MAXPGPATH];
-	char		env_path[MAXPGPATH + sizeof("PGSYSCONFDIR=")];	/* longer than
-																 * PGLOCALEDIR */
-
-	/* don't set LC_ALL in the backend */
-	if (strcmp(app, "postgres") != 0)
-		setlocale(LC_ALL, "");
-
-	if (find_my_exec(argv0, my_exec_path) < 0)
-		return;
-
-#ifdef ENABLE_NLS
-	get_locale_path(my_exec_path, path);
-	bindtextdomain(app, path);
-	textdomain(app);
-
-	if (getenv("PGLOCALEDIR") == NULL)
-	{
-		/* set for libpq to use */
-		snprintf(env_path, sizeof(env_path), "PGLOCALEDIR=%s", path);
-		canonicalize_path(env_path + 12);
-		putenv(strdup(env_path));
-	}
-#endif
-
-	if (getenv("PGSYSCONFDIR") == NULL)
-	{
-		get_etc_path(my_exec_path, path);
-
-		/* set for libpq to use */
-		snprintf(env_path, sizeof(env_path), "PGSYSCONFDIR=%s", path);
-		canonicalize_path(env_path + 13);
-		putenv(strdup(env_path));
-	}
-}
-
-
 /*
  *	trim_directory
  *
